/*
 * wpa_supplicant/hostapd: TLSv1 common routines
 * Copyright (c) 2006, Jouni Malinen <jkmaline@cc.hut.fi>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Alternatively, this software may be distributed under the terms of BSD
 * license.
 *
 * See README and COPYING for more details.
 */

#include "includes.h"
#include <tomcrypt.h>

#include "common.h"
#include "md5.h"
#include "sha1.h"
#include "rc4.h"
#include "tlsv1_common.h"


/*
 * TODO:
 * RFC 2246 Section 9: Mandatory to implement TLS_DHE_DSS_WITH_3DES_EDE_CBC_SHA
 * Add support for commonly used cipher suites; don't bother with exportable
 * suites.
 */ 

static const struct tls_cipher_suite tls_cipher_suites[] = {
	{ TLS_NULL_WITH_NULL_NULL, TLS_KEY_X_NULL, TLS_CIPHER_NULL,
	  TLS_HASH_NULL },
	{ TLS_RSA_WITH_RC4_128_MD5, TLS_KEY_X_RSA, TLS_CIPHER_RC4_128,
	  TLS_HASH_MD5 },
	{ TLS_RSA_WITH_RC4_128_SHA, TLS_KEY_X_RSA, TLS_CIPHER_RC4_128,
	  TLS_HASH_SHA },
	{ TLS_RSA_WITH_DES_CBC_SHA, TLS_KEY_X_RSA, TLS_CIPHER_DES_CBC,
	  TLS_HASH_SHA },
	{ TLS_RSA_WITH_3DES_EDE_CBC_SHA, TLS_KEY_X_RSA,
	  TLS_CIPHER_3DES_EDE_CBC, TLS_HASH_SHA },
	{ TLS_RSA_WITH_AES_128_CBC_SHA, TLS_KEY_X_RSA, TLS_CIPHER_AES_128_CBC,
	  TLS_HASH_SHA },
	{ TLS_RSA_WITH_AES_256_CBC_SHA, TLS_KEY_X_RSA, TLS_CIPHER_AES_256_CBC,
	  TLS_HASH_SHA }
	/* TODO: add support for more cipher suites */
};

#define NUM_ELEMS(a) (sizeof(a) / sizeof((a)[0]))
#define NUM_TLS_CIPHER_SUITES NUM_ELEMS(tls_cipher_suites)


static const struct tls_cipher_data tls_ciphers[] = {
	{ TLS_CIPHER_NULL,         TLS_CIPHER_STREAM,  0,  0,  0, NULL },
	{ TLS_CIPHER_IDEA_CBC,     TLS_CIPHER_BLOCK,  16, 16,  8, NULL },
	{ TLS_CIPHER_RC2_CBC_40,   TLS_CIPHER_BLOCK,   5, 16,  0, "rc2" },
	{ TLS_CIPHER_RC4_40,       TLS_CIPHER_STREAM,  5, 16,  0, NULL },
	{ TLS_CIPHER_RC4_128,      TLS_CIPHER_STREAM, 16, 16,  0, NULL },
	{ TLS_CIPHER_DES40_CBC,    TLS_CIPHER_BLOCK,   5,  8,  8, "des" },
	{ TLS_CIPHER_DES_CBC,      TLS_CIPHER_BLOCK,   8,  8,  8, "des" },
	{ TLS_CIPHER_3DES_EDE_CBC, TLS_CIPHER_BLOCK,  24, 24,  8, "3des" },
	{ TLS_CIPHER_AES_128_CBC,  TLS_CIPHER_BLOCK,  16, 16, 16, "aes" },
	{ TLS_CIPHER_AES_256_CBC,  TLS_CIPHER_BLOCK,  32, 32, 16, "aes" }
};

#define NUM_TLS_CIPHER_DATA NUM_ELEMS(tls_ciphers)


const struct tls_cipher_suite * tls_get_cipher_suite(u16 suite)
{
	size_t i;
	for (i = 0; i < NUM_TLS_CIPHER_SUITES; i++)
		if (tls_cipher_suites[i].suite == suite)
			return &tls_cipher_suites[i];
	return NULL;
}


static const struct tls_cipher_data * tls_get_cipher_data(tls_cipher cipher)
{
	size_t i;
	for (i = 0; i < NUM_TLS_CIPHER_DATA; i++)
		if (tls_ciphers[i].cipher == cipher)
			return &tls_ciphers[i];
	return NULL;
}


#ifndef CONFIG_NO_STDOUT_DEBUG
static const char * tls_asn1_type_str(int type)
{
	switch (type) {
	case LTC_ASN1_EOL:
		return "EOL";
	case LTC_ASN1_INTEGER:
		return "INTEGER";
	case LTC_ASN1_SHORT_INTEGER:
		return "SHORT_INTEGER";
	case LTC_ASN1_BIT_STRING:
		return "BIT_STRING";
	case LTC_ASN1_OCTET_STRING:
		return "OCTET_STRING";
	case LTC_ASN1_NULL:
		return "NULL";
	case LTC_ASN1_OBJECT_IDENTIFIER:
		return "OBJECT_IDENTIFIER";
	case LTC_ASN1_IA5_STRING:
		return "IA5_STRING";
	case LTC_ASN1_PRINTABLE_STRING:
		return "PRINTABLE_STRING";
	case LTC_ASN1_UTCTIME:
		return "UTCTIME";
	case LTC_ASN1_CHOICE:
		return "CHOICE";
	case LTC_ASN1_SEQUENCE:
		return "SEQUENCE";
	case LTC_ASN1_SET:
		return "SET";
	case LTC_ASN1_SETOF:
		return "SETOF";
	}
	return "?";
}
#endif /* CONFIG_NO_STDOUT_DEBUG */


static void tls_asn1_dump(ltc_asn1_list *asn1, unsigned int level)
{
	char prefix[100], buf[100], *pos;
	size_t i;
	unsigned long *oid;

	memset(prefix, ' ', sizeof(prefix));
	if (level * 4 >= sizeof(prefix) - 1)
		prefix[sizeof(prefix) - 1] = '\0';
	else
		prefix[level * 4] = '\0';

	while (asn1) {
		wpa_printf(MSG_MSGDUMP, "ASN.1:%s type=%d (%s) size=%lu",
			   prefix, asn1->type, tls_asn1_type_str(asn1->type),
			   (unsigned long) asn1->size);
		switch (asn1->type) {
		case LTC_ASN1_PRINTABLE_STRING:
		case LTC_ASN1_IA5_STRING:
			wpa_hexdump_ascii(MSG_MSGDUMP,
					  "ASN.1: ASCII dump",
					  asn1->data, asn1->size);
		case LTC_ASN1_INTEGER:
		case LTC_ASN1_SHORT_INTEGER:
		case LTC_ASN1_UTCTIME:
			wpa_hexdump(MSG_MSGDUMP, "ASN.1: data",
				    asn1->data, asn1->size);
			break;
		case LTC_ASN1_OBJECT_IDENTIFIER:
			pos = buf;
			oid = asn1->data;
			for (i = 0; i < asn1->size; i++)
				pos += snprintf(pos, buf + sizeof(buf) - pos,
						".%lu", oid[i]);
			wpa_printf(MSG_MSGDUMP, "ASN.1:%s   OID %s",
				   prefix, buf);
			break;
		default:
			break;
		}

		if (asn1->child)
			tls_asn1_dump(asn1->child, level + 1);
		asn1 = asn1->next;
	}
}


static int tls_asn1_oid_rsa_encryption(unsigned long *oid, size_t len)
{
	if (len != 7)
		return 0;
	return oid[0] == 1 && oid[1] == 2 && oid[2] == 840 &&
		oid[3] == 113549 && oid[4] == 1 && oid[5] == 1 && oid[6] == 1;
}


static u8 * tls_asn1_bit_to_byte(const u8 *bits, size_t num_bits, size_t *len)
{
	u8 *buf, val;
	size_t i, j;

	*len = num_bits / 8;
	buf = malloc(*len);
	if (buf == NULL)
		return NULL;

	for (i = 0; i < *len; i++) {
		val = 0;
		for (j = 0; j < 8; j++)
			val = (val << 1) | bits[i * 8 + j];
		buf[i] = val;
	}

	return buf;
}


static u8 * tls_asn1_get_public_key(ltc_asn1_list *asn1, size_t *len)
{
	u8 *res;

	while (asn1) {
		if (asn1->type == LTC_ASN1_SEQUENCE &&
		    asn1->child &&
		    asn1->child->type == LTC_ASN1_OBJECT_IDENTIFIER &&
		    tls_asn1_oid_rsa_encryption(asn1->child->data,
						asn1->child->size) &&
		    asn1->next && asn1->next->type == LTC_ASN1_BIT_STRING) {
			return tls_asn1_bit_to_byte(asn1->next->data,
						    asn1->next->size, len);
		}
		    
		if (asn1->child) {
			res = tls_asn1_get_public_key(asn1->child, len);
			if (res)
				return res;
		}
		asn1 = asn1->next;
	}

	return NULL;
}


int tls_parse_cert(const u8 *buf, size_t len, rsa_key *rsa)
{
	int res;
	ltc_asn1_list *asn1;
	unsigned long ilen;
	u8 *public_key;
	size_t public_key_len;

	wpa_hexdump(MSG_MSGDUMP, "TLSv1: Parse ASN.1 DER certificate",
		    buf, len);
	ilen = len;
	res = der_decode_sequence_flexi(buf, &ilen, &asn1);
	if (res != CRYPT_OK) {
		wpa_printf(MSG_DEBUG, "TLSv1: ASN.1 parsing failed: %s",
			   error_to_string(res));
		return -1;
	}

	if (ilen != len) {
		wpa_printf(MSG_DEBUG, "TLSv1: ASN.1 decoding did not process "
			   "full data");
		der_sequence_free(asn1);
		return -1;
	}

	tls_asn1_dump(asn1, 0);

	/* TODO
	 * verify key usage (must allow encryption)
	 *
	 * All certificate profiles, key and cryptographic formats are
	 * defined by the IETF PKIX working group [PKIX]. When a key
	 * usage extension is present, the digitalSignature bit must be
	 * set for the key to be eligible for signing, as described
	 * above, and the keyEncipherment bit must be present to allow
	 * encryption, as described above. The keyAgreement bit must be
	 * set on Diffie-Hellman certificates. (PKIX: RFC 3280)
	 */

	public_key = tls_asn1_get_public_key(asn1, &public_key_len);
	if (public_key == NULL) {
		wpa_printf(MSG_ERROR, "TLSv1: No public key found from the "
			   "certificate");
		der_sequence_free(asn1);
		return -1;
	}

	wpa_hexdump(MSG_MSGDUMP, "TLSv1: Public key",
		    public_key, public_key_len);
	res = rsa_import(public_key, public_key_len, rsa);
	free(public_key);

	if (res != CRYPT_OK) {
		wpa_printf(MSG_ERROR, "TLSv1: Failed to parse "
			   "server certificate (res=%d '%s')",
			   res, error_to_string(res));
		der_sequence_free(asn1);
		return -1;
	}

	der_sequence_free(asn1);
	return 0;
}


int tlsv1_record_set_cipher_suite(struct tlsv1_record_layer *rl,
				  u16 cipher_suite)
{
	const struct tls_cipher_suite *suite;
	const struct tls_cipher_data *data;

	wpa_printf(MSG_DEBUG, "TLSv1: Selected cipher suite: 0x%04x",
		   cipher_suite);
	rl->cipher_suite = cipher_suite;

	suite = tls_get_cipher_suite(cipher_suite);
	if (suite) {
		if (suite->hash == TLS_HASH_MD5) {
			rl->hash_idx = find_hash("md5");
			rl->hash_size = MD5_MAC_LEN;
		} else if (suite->hash == TLS_HASH_SHA) {
			rl->hash_idx = find_hash("sha1");
			rl->hash_size = SHA1_MAC_LEN;
		}

		rl->use_rc4_cipher = suite->cipher == TLS_CIPHER_RC4_128;
		rl->rc4_read_len = rl->rc4_write_len = 0;
		data = tls_get_cipher_data(suite->cipher);
		if (data == NULL)
			return -1;

		rl->key_material_len = data->key_material;
		rl->iv_size = data->block_size;
		if (data->name)
			rl->cipher_idx = find_cipher(data->name);
		else
			rl->cipher_idx = -2;

		if (rl->hash_idx < 0) {
			wpa_printf(MSG_DEBUG, "TLSv1: Could not select hash "
				   "algorithm");
			return -1;
		}
		if (rl->cipher_idx == -1) {
			wpa_printf(MSG_DEBUG, "TLSv1: Could not select cipher "
				   "algorithm");
			return -1;
		}
	} else {
		rl->cipher_idx = -2;
	}

	return 0;
}


int tlsv1_record_change_write_cipher(struct tlsv1_record_layer *rl)
{
	int res;

	wpa_printf(MSG_DEBUG, "TLSv1: Record Layer - New write cipher suite "
		   "0x%04x", rl->cipher_suite);
	rl->write_cipher_suite = rl->cipher_suite;
	memset(rl->write_seq_num, 0, TLS_SEQ_NUM_LEN);

	if (rl->write_cbc_initialized) {
		cbc_done(&rl->write_cbc);
		rl->write_cbc_initialized = 0;
	}
	if (rl->cipher_idx >= 0) {
		res = cbc_start(rl->cipher_idx, rl->write_iv, rl->write_key,
				rl->key_material_len, 0, &rl->write_cbc);
		if (res != CRYPT_OK) {
			wpa_printf(MSG_DEBUG, "TLSv1: Cipher start failed: %s",
				   error_to_string(res));
			return -1;
		}
		rl->write_cbc_initialized = 1;
	}

	return 0;
}


int tlsv1_record_change_read_cipher(struct tlsv1_record_layer *rl)
{
	int res;

	wpa_printf(MSG_DEBUG, "TLSv1: Record Layer - New read cipher suite "
		   "0x%04x", rl->cipher_suite);
	rl->read_cipher_suite = rl->cipher_suite;
	memset(rl->read_seq_num, 0, TLS_SEQ_NUM_LEN);

	if (rl->read_cbc_initialized) {
		cbc_done(&rl->read_cbc);
		rl->read_cbc_initialized = 0;
	}
	if (rl->cipher_idx >= 0) {
		res = cbc_start(rl->cipher_idx, rl->read_iv, rl->read_key,
				rl->key_material_len, 0, &rl->read_cbc);
		if (res != CRYPT_OK) {
			wpa_printf(MSG_DEBUG, "TLSv1: Cipher start failed: %s",
				   error_to_string(res));
			return -1;
		}
		rl->read_cbc_initialized = 1;
	}

	return 0;
}


int tlsv1_record_send(struct tlsv1_record_layer *rl, u8 content_type, u8 *buf,
		      size_t buf_size, size_t payload_len, size_t *out_len)
{
	u8 *pos, *ct_start, *length, *payload;
	hmac_state hmac;
	unsigned long clen;
	int res;

	pos = buf;
	/* ContentType type */
	ct_start = pos;
	*pos++ = content_type;
	/* ProtocolVersion version */
	WPA_PUT_BE16(pos, TLS_VERSION);
	pos += 2;
	/* uint16 length */
	length = pos;
	WPA_PUT_BE16(length, payload_len);
	pos += 2;

	/* opaque fragment[TLSPlaintext.length] */
	payload = pos;
	pos += payload_len;

	if (rl->write_cipher_suite != TLS_NULL_WITH_NULL_NULL) {
		hmac_init(&hmac, rl->hash_idx, rl->write_mac_secret,
			  rl->hash_size);
		hmac_process(&hmac, rl->write_seq_num, TLS_SEQ_NUM_LEN);
		/* type + version + length + fragment */
		hmac_process(&hmac, ct_start, pos - ct_start);
		clen = buf + buf_size - pos;
		if (clen < rl->hash_size) {
			wpa_printf(MSG_DEBUG, "TLSv1: Record Layer - Not "
				   "enough room for MAC");
			return -1;
		}

		hmac_done(&hmac, pos, &clen);
		wpa_hexdump(MSG_MSGDUMP, "TLSv1: Record Layer - Write HMAC",
			    pos, clen);
		pos += clen;
		if (rl->use_rc4_cipher) {
			rc4_skip(rl->write_key, rl->key_material_len,
				 rl->rc4_write_len, payload, pos - payload);
			rl->rc4_write_len += pos - payload;
		} else if (rl->iv_size) {
			size_t len = pos - payload;
			size_t pad;
			pad = (len + 1) % rl->iv_size;
			if (pad)
				pad = rl->iv_size - pad;
			if (pos + pad + 1 > buf + buf_size) {
				wpa_printf(MSG_DEBUG, "TLSv1: No room for "
					   "block cipher padding");
				return -1;
			}
			memset(pos, pad, pad + 1);
			pos += pad + 1;

			res = cbc_encrypt(payload, payload, pos - payload,
					  &rl->write_cbc);
			if (res != CRYPT_OK) {
				wpa_printf(MSG_DEBUG, "TLSv1: CBC encryption "
					   "failed: %s", error_to_string(res));
				return -1;
			}
		} else
			return -1;
	}

	WPA_PUT_BE16(length, pos - length - 2);
	inc_byte_array(rl->write_seq_num, TLS_SEQ_NUM_LEN);

	*out_len = pos - buf;

	return 0;
}


int tlsv1_record_receive(struct tlsv1_record_layer *rl,
			 const u8 *in_data, size_t in_len,
			 u8 *out_data, size_t *out_len, u8 *alert)
{
	int res;
	size_t i, rlen;
	unsigned long hlen;
	u8 padlen;
	hmac_state hmac;
	u8 len[2], hash[100];

	wpa_hexdump(MSG_MSGDUMP, "TLSv1: Record Layer - Received",
		    in_data, in_len);

	if (in_len < TLS_RECORD_HEADER_LEN) {
		wpa_printf(MSG_DEBUG, "TLSv1: Too short record (in_len=%lu)",
			   (unsigned long) in_len);
		*alert = TLS_ALERT_DECODE_ERROR;
		return -1;
	}

	wpa_printf(MSG_DEBUG, "TLSv1: Received content type %d version %d.%d "
		   "length %d", in_data[0], in_data[1], in_data[2],
		   WPA_GET_BE16(in_data + 3));

	if (in_data[0] != TLS_CONTENT_TYPE_HANDSHAKE &&
	    in_data[0] != TLS_CONTENT_TYPE_CHANGE_CIPHER_SPEC &&
	    in_data[0] != TLS_CONTENT_TYPE_APPLICATION_DATA) {
		wpa_printf(MSG_DEBUG, "TLSv1: Unexpected content type 0x%x",
			   in_data[0]);
		*alert = TLS_ALERT_UNEXPECTED_MESSAGE;
		return -1;
	}

	if (WPA_GET_BE16(in_data + 1) != TLS_VERSION) {
		wpa_printf(MSG_DEBUG, "TLSv1: Unexpected protocol version "
			   "%d.%d", in_data[1], in_data[2]);
		*alert = TLS_ALERT_PROTOCOL_VERSION;
		return -1;
	}

	rlen = WPA_GET_BE16(in_data + 3);
	in_data += TLS_RECORD_HEADER_LEN;
	in_len -= TLS_RECORD_HEADER_LEN;

	if (rlen > in_len) {
		wpa_printf(MSG_DEBUG, "TLSv1: Not all record data included "
			   "(rlen=%lu > in_len=%lu)",
			   (unsigned long) rlen, (unsigned long) in_len);
		*alert = TLS_ALERT_DECODE_ERROR;
		return -1;
	}

	in_len = rlen;

	if (*out_len < in_len) {
		wpa_printf(MSG_DEBUG, "TLSv1: Not enough output buffer for "
			   "processing received record");
		*alert = TLS_ALERT_INTERNAL_ERROR;
		return -1;
	}

	memcpy(out_data, in_data, in_len);
	*out_len = in_len;

	if (rl->read_cipher_suite != TLS_NULL_WITH_NULL_NULL) {
		if (rl->use_rc4_cipher) {
			rc4_skip(rl->read_key, rl->key_material_len,
				 rl->rc4_read_len, out_data, in_len);
			rl->rc4_read_len += in_len;
		} else if (rl->iv_size) {
			res = cbc_decrypt(out_data, out_data, in_len,
					  &rl->read_cbc);
			if (res != CRYPT_OK) {
				wpa_printf(MSG_DEBUG, "TLSv1: CBC decryption "
					   "failed: %s", error_to_string(res));
				*alert = TLS_ALERT_DECRYPTION_FAILED;
				return -1;
			}

			if (in_len == 0) {
				wpa_printf(MSG_DEBUG, "TLSv1: Too short record"
					   " (no pad)");
				*alert = TLS_ALERT_DECODE_ERROR;
				return -1;
			}
			padlen = out_data[in_len - 1];
			if (padlen >= in_len) {
				wpa_printf(MSG_DEBUG, "TLSv1: Incorrect pad "
					   "length (%u, in_len=%lu) in "
					   "received record",
					   padlen, (unsigned long) in_len);
				*alert = TLS_ALERT_DECRYPTION_FAILED;
				return -1;
			}
			for (i = in_len - padlen; i < in_len; i++) {
				if (out_data[i] != padlen) {
					wpa_hexdump(MSG_DEBUG,
						    "TLSv1: Invalid pad in "
						    "received record",
						    out_data + in_len - padlen,
						    padlen);
					*alert = TLS_ALERT_DECRYPTION_FAILED;
					return -1;
				}
			}

			*out_len -= padlen + 1;
		} else {
			wpa_printf(MSG_DEBUG, "TLSv1: No supported cipher for "
				   "processing the received record");
			*alert = TLS_ALERT_INTERNAL_ERROR;
			return -1;
		}

		wpa_hexdump(MSG_MSGDUMP,
			    "TLSv1: Record Layer - Decrypted data",
			    out_data, in_len);

		if (*out_len < rl->hash_size) {
			wpa_printf(MSG_DEBUG, "TLSv1: Too short record; no "
				   "hash value");
			*alert = TLS_ALERT_INTERNAL_ERROR;
			return -1;
		}

		*out_len -= rl->hash_size;

		hmac_init(&hmac, rl->hash_idx, rl->read_mac_secret,
			  rl->hash_size);
		hmac_process(&hmac, rl->read_seq_num, TLS_SEQ_NUM_LEN);
		/* type + version + length + fragment */
		hmac_process(&hmac, in_data - TLS_RECORD_HEADER_LEN, 3);
		WPA_PUT_BE16(len, *out_len);
		hmac_process(&hmac, len, 2);
		hmac_process(&hmac, out_data, *out_len);
		hlen = sizeof(hash);
		hmac_done(&hmac, hash, &hlen);
		if (hlen != rl->hash_size ||
		    memcmp(hash, out_data + *out_len, hlen) != 0) {
			wpa_printf(MSG_DEBUG, "TLSv1: Invalid HMAC value in "
				   "received message");
			*alert = TLS_ALERT_BAD_RECORD_MAC;
			return -1;
		}
	}

	inc_byte_array(rl->read_seq_num, TLS_SEQ_NUM_LEN);

	return 0;
}
